PACKAGE_NAME = 'marrs'

FRIDA_SERVER_VERSION = '14.2.18'
FRIDA_SERVER_FILENAME = 'frida-server-{version}-android-{arch}'
FRIDA_SERVER_DOWNLOAD_URL = 'https://github.com/frida/frida/releases/download/{version}/{filename}.xz'
FRIDA_SERVER_REMOTE_PATH = '/data/local/tmp/frida-server-{version}-{arch}'

STORAGE_TMP_DIR_NAME = 'tmp'
